import click


@click.command(context_settings=dict(help_option_names=['-h', '--help']))
@click.argument('paths', nargs=-1)
@click.option('-n', '--name-regex', metavar='regex',
              help=r'a regex used to infer DataBlock names from paths [fallback: \d+]')
@click.option('--strict', is_flag=True,
              help='immediately fail if a matched path cannot be read')
@click.option('--lazy', is_flag=True, default=True,
              help='read data lazily (if possible)')
@click.version_option()
def cli(paths, name_regex, strict, lazy):
    """
    naaf command line interface.

    Opens files and lands in an interactive ipython shell with data loaded as `data`

    PATHS: any number of files [default='.']

    EXAMPLES:

    Open a .star file as particles:

        naaf particles.star

    Open particles and images from a directory:

        naaf /dir/with/star_and_mrc_files/

    Match files such as MyProtein_10.star and MyProtein_001.mrc,
    and name the respective data objects Protein_10 and Protein_001:

        naaf /path/to/dir/MyProtein* -n 'Protein_\d+'
    """
    if not paths:
        paths = ['./*']

    import naaf
    from IPython.terminal.embed import InteractiveShellEmbed

    data = naaf.read(
        *paths,  # noqa: F841
         name_regex=name_regex,
         strict=strict,
         lazy=lazy,
     )

    # set up ipython shell nicely
    banner = '=== naaf ==='
    sh = InteractiveShellEmbed(banner2=banner)
    sh.push('data')
    sh()



# euler angle definitions are as expected by scipy Rotation objects
# uppercase is intrinsic, lowercase extrinsic


class Naaf:
    COORD_HEADERS = ["x", "y", "z"]
    ROT_HEADER = "rot"
    ALL_HEADERS = COORD_HEADERS + [ROT_HEADER]


class Relion:
    COORD_HEADERS = [f"rlnCoordinate{axis}" for axis in "XYZ"]
    EULER_HEADERS = {
        3: [f"rlnAngle{angle}" for angle in ("Rot", "Tilt", "Psi")],
        2: ["rlnAnglePsi"],
    }
    SHIFT_HEADERS = {
        "3.0": [f"rlnOrigin{axis}" for axis in "XYZ"],
        "3.1": [f"rlnOrigin{axis}Angst" for axis in "XYZ"],
    }
    PIXEL_SIZE_HEADERS = {"3.0": ["rlnDetectorPixelSize"], "3.1": ["rlnImagePixelSize"]}
    MICROGRAPH_NAME_HEADER = "rlnMicrographName"
    OPTICS_GROUP_HEADER = "rlnOpticsGroup"
    ALL_HEADERS = (
        COORD_HEADERS
        + EULER_HEADERS[3]
        + SHIFT_HEADERS["3.0"]
        + SHIFT_HEADERS["3.1"]
        + PIXEL_SIZE_HEADERS["3.0"]
        + PIXEL_SIZE_HEADERS["3.1"]
        + [MICROGRAPH_NAME_HEADER]
        + [OPTICS_GROUP_HEADER]
    )

    EULER = "ZYZ"
    INPLANE = "Z"


class Dynamo:
    COORD_HEADERS = ["x", "y", "z"]
    SHIFT_HEADERS = ["dx", "dy", "dz"]
    EULER_HEADERS = {
        3: ["tdrot", "tilt", "narot"],
        2: ["tilt"],  # TODO: 2d column name might be wrong!
    }
    ALL_HEADERS = COORD_HEADERS + SHIFT_HEADERS + EULER_HEADERS[3]

    EULER = "zxz"
    INPLANE = "z"



import numpy as np
import mrcfile

from naaf.reading.mrc import read_mrc
from naaf.data import Image
from naaf.utils.testing import assert_data_equal


def test_read_mrc(tmp_path):
    file_path = tmp_path / 'test.mrc'
    data = np.ones((3, 3, 3), dtype=np.float32)
    mrcfile.new(str(file_path), data)
    image = read_mrc(file_path, name_regex=r'\w+', lazy=False)

    expected = Image(
        data=data,
        name='test',
    )

    assert_data_equal(image, expected)



import dynamotable
import numpy as np
import pandas as pd
from scipy.spatial.transform import Rotation

from ..data import Particles
from ..utils.constants import Dynamo, Naaf
from ..utils.generic import guess_name


def name_from_volume(volume_identifier, name_regex=None):
    """Generate ParticleBlock name from volume identifier from dataframe"""
    if isinstance(volume_identifier, int):
        return str(volume_identifier)
    elif isinstance(volume_identifier, str):
        return guess_name(volume_identifier, name_regex)


def read_tbl(table_path, table_map_file=None, name_regex=None, **kwargs):
    """
    Read particles from a dynamo format table file
    """
    df = dynamotable.read(table_path, table_map_file)

    split_on = "tomo"
    if "tomo_file" in df.columns:
        split_on = "tomo_file"

    if Dynamo.COORD_HEADERS[-1] in df.columns:
        dim = 3
    else:
        dim = 2

    particles = []
    for volume, df_volume in df.groupby(split_on):
        # drop global index to prevent issues with concatenation and similar
        df_volume = df_volume.reset_index(drop=True)

        name = name_from_volume(volume, name_regex)
        coords = np.asarray(df_volume[Dynamo.COORD_HEADERS[:dim]], dtype=float)
        shifts = np.asarray(df_volume.get(Dynamo.SHIFT_HEADERS[:dim], 0), dtype=float)
        coords += shifts

        eulers = np.asarray(df_volume.get(Dynamo.EULER_HEADERS[dim], 0), dtype=float)
        if dim == 3:
            rot = Rotation.from_euler(Dynamo.EULER, eulers, degrees=True)
        else:
            rot = Rotation.from_euler(Dynamo.INPLANE, eulers, degrees=True)

        # we want the inverse, which when applied to basis vectors it gives us the particle orientation
        rot = rot.inv()

        data = pd.DataFrame()
        data[Naaf.COORD_HEADERS] = coords
        data[Naaf.ROT_HEADER] = np.asarray(rot)

        particles.append(
            Particles(
                data=data,
                name=name,
            )
        )

    return particles



from pathlib import Path
from typing import Optional, Union

import dask.array as da
import numpy as np
import pandas as pd
from pydantic import BaseModel, validator


class Data(BaseModel):
    class Config:
        arbitrary_types_allowed = True

    name: Optional[str] = None
    source: Optional[Path] = None
    pixel_size: Optional[float] = None

    @validator("pixel_size")
    def _validate_pixel_size(cls, v):
        if v is not None:
            # TODO: for now just make everything 3D
            return np.broadcast_to(v, (3,))


class Particles(Data):
    data: pd.DataFrame


class Image(Data):
    data: Union[np.ndarray, da.Array]
    stack: bool = False



import numpy as np
import pandas as pd
import dynamotable
from scipy.spatial.transform import Rotation

from naaf.reading.tbl import read_tbl
from naaf.data import Particles
from naaf.utils.constants import Naaf, Dynamo
from naaf.utils.testing import assert_data_equal


def test_read_tbl(tmp_path):
    df = pd.DataFrame({
        'x': [1, 1],
        'y': [2, 2],
        'z': [3, 3],
        'dx': [0.1, 0.1],
        'dy': [0.2, 0.2],
        'dz': [0.3, 0.3],
        'tdrot': [0, 0],
        'tilt': [0, 90],
        'narot': [90, 0],
        'tomo': [0, 1],
    })
    file_path = tmp_path / 'test.tbl'
    dynamotable.write(df, file_path)

    particles = read_tbl(file_path, name_regex=r'\w')
    part = particles[0]

    expected_data = pd.DataFrame()
    expected_data[Naaf.COORD_HEADERS] = np.array([[1.1, 2.2, 3.3]])
    expected_data[Naaf.ROT_HEADER] = Rotation.from_euler(Dynamo.EULER, [[0, 0, 90]], degrees=True).inv()

    expected = Particles(
        data=expected_data,
        name='0',
    )

    assert_data_equal(part, expected)



from pathlib import Path

from ..utils.generic import ParseError, listify
from .box import read_box
from .cbox import read_cbox
from .em import read_em
from .mrc import read_mrc
from .star import read_star
from .tbl import read_tbl

# a mapping of file extensions to readers, tuple map to tuples (don't forget trailing comma!):
#   - multiple extensions values in the keys use the same readers
#   - multiple readers are called in order from highest to lowers in case previous ones fail
# TODO: put this directly in the readers to make it plug and play?
readers = {
    (".star",): (read_star,),
    (".mrc", ".mrcs", ".st", ".map"): (read_mrc,),
    (".em",): (read_em,),
    (".tbl",): (read_tbl,),
    (".box",): (read_box,),
    (".cbox",): (read_cbox,),
}

known_formats = [ext for formats in readers for ext in formats]


def read_file(file_path, **kwargs):
    """
    read a single file with the appropriate parser and return a list
    """
    for ext, funcs in readers.items():
        if file_path.suffix in ext:
            for func in funcs:
                try:
                    data = listify(func(file_path, **kwargs))
                    for d in data:
                        d.source = file_path
                    return data
                except ParseError:
                    # this will be raised by individual readers when the file can't be read.
                    # Keep trying until all options are exhausted
                    continue
    raise ParseError(f"could not read {file_path}")


def filter_readable(paths):
    for path in paths:
        path = Path(path)
        if path.is_dir():
            yield from filter_readable(path.iterdir())
        if path.is_file() and path.suffix in known_formats:
            yield path


def read(
    *paths,
    name_regex=None,
    strict=False,
    lazy=True,
    **kwargs,
):
    r"""
    Read any number of paths.

    name_regex: a regex used to infer names from paths or micrograph names. For example:
                'Protein_\d+' will match 'MyProtein_10.star' and 'MyProtein_001.mrc'
                and name the respective DataBlocks 'Protein_10' and 'Protein_01'
    strict:     if set to true, immediately fail if a matched filename cannot be read
    lazy:       read data lazily (if possible)
    """
    data = []
    for file in filter_readable(paths):
        try:
            data.extend(read_file(file, name_regex=name_regex, lazy=lazy, **kwargs))
        except ParseError:
            if strict:
                raise
    if not data and strict:
        raise ParseError(f"could not read any data from {paths}")

    return data



import numpy as np
import emfile

from naaf.reading.em import read_em
from naaf.data import Image
from naaf.utils.testing import assert_data_equal


def test_read_em(tmp_path):
    file_path = tmp_path / 'test.em'
    data = np.ones((3, 3, 3))
    emfile.write(str(file_path), data)
    image = read_em(file_path, name_regex=r'\w+', lazy=False)
    
    expected = Image(
        data=data,
        name='test',
    )

    assert_data_equal(image, expected)



import numpy as np
import pandas as pd
import starfile
from scipy.spatial.transform import Rotation

from naaf.reading.star import read_star
from naaf.data import Particles
from naaf.utils.constants import Naaf, Relion
from naaf.utils.testing import assert_data_equal


def test_read_relion30_3d(tmp_path):
    df = pd.DataFrame({
        'rlnCoordinateX': [1, 1],
        'rlnCoordinateY': [2, 2],
        'rlnCoordinateZ': [3, 3],
        'rlnOriginX': [0.1, 0.1],
        'rlnOriginY': [0.2, 0.2],
        'rlnOriginZ': [0.3, 0.3],
        'rlnAngleRot': [0, 0],
        'rlnAngleTilt': [0, 90],
        'rlnAnglePsi': [90, 0],
        'rlnMicrographName': ['a', 'b'],
        'feature': ['x', 'y'],
    })
    file_path = tmp_path / 'test.star'
    starfile.write(df, file_path)

    particles = read_star(file_path, name_regex=r'\w')
    part = particles[0]

    expected_data = pd.DataFrame()
    expected_data[Naaf.COORD_HEADERS] = np.array([[0.9, 1.8, 2.7]])
    expected_data[Naaf.ROT_HEADER] = Rotation.from_euler(Relion.EULER, [[0, 0, 90]], degrees=True).inv()
    expected_data['feature'] = 'x'

    expected = Particles(
        data=expected_data,
        name='a',
    )
    assert_data_equal(part, expected)



def test_read_relion31_3d(tmp_path):
    df_optics = pd.DataFrame({
        'rlnOpticsGroup': [1],
        'rlnImagePixelSize': [10],
    })
    df_particles = pd.DataFrame({
        'rlnCoordinateX': [1, 1],
        'rlnCoordinateY': [2, 2],
        'rlnCoordinateZ': [3, 3],
        'rlnOriginXAngst': [0.1, 0.1],
        'rlnOriginYAngst': [0.2, 0.2],
        'rlnOriginZAngst': [0.3, 0.3],
        'rlnAngleRot': [0, 0],
        'rlnAngleTilt': [0, 90],
        'rlnAnglePsi': [90, 0],
        'rlnMicrographName': ['a', 'b'],
        'feature': ['x', 'y'],
        'rlnOpticsGroup': [1, 1],
    })
    data_dict = {'optics': df_optics, 'particles': df_particles}
    file_path = tmp_path / 'test.star'
    starfile.write(data_dict, file_path)

    particles = read_star(file_path, name_regex=r'\w')
    part = particles[0]

    expected_data = pd.DataFrame()
    expected_data[Naaf.COORD_HEADERS] = np.array([[0.99, 1.98, 2.97]])
    expected_data[Naaf.ROT_HEADER] = Rotation.from_euler(Relion.EULER, [[0, 0, 90]], degrees=True).inv()
    expected_data['feature'] = 'x'

    expected = Particles(
        data=expected_data,
        pixel_size=10,
        name='a',
    )
    assert_data_equal(part, expected)



import numpy as np
import pandas as pd
from scipy.spatial.transform import Rotation

from naaf.writing.tbl import write_tbl
from naaf.data import Particles
from naaf.utils.constants import Naaf


def test_write_tbl(tmp_path):
    file_path = tmp_path / 'test.tbl'
    data = pd.DataFrame()
    data[Naaf.COORD_HEADERS] = np.ones((2, 3))
    data[Naaf.ROT_HEADER] = np.asarray(Rotation.identity(2))
    data['feature'] = ['x', 'y']

    particle = Particles(
        data=data,
        name='test',
    )

    write_tbl(particle, file_path)



import mrcfile


def write_mrc(image, file_path, overwrite=False):
    """
    write an image to disk as an .mrc file
    """
    if not file_path.endswith('.mrc'):
        file_path = file_path + '.mrc'
    # TODO: pass pixel size and other stuff to header
    mrcfile.new(file_path, image.data, overwrite=overwrite)



import pandas as pd
from scipy.spatial.transform import Rotation
import dynamotable

from ..utils.constants import Naaf, Dynamo


def write_tbl(particles, file_path):
    """
    write particle data to disk as a dynamo .tbl file
    """
    df = pd.DataFrame()
    df[Dynamo.COORD_HEADERS] = particles.data[Naaf.COORD_HEADERS]
    rot = Rotation.concatenate(particles.data[Naaf.ROT_HEADER])
    # we use the inverse rotation in naaf
    df[Dynamo.EULER_HEADERS[3]] = rot.inv().as_euler(Dynamo.EULER, degrees=True)

    if not str(file_path).endswith('.star'):
        file_path = str(file_path) + '.star'
    dynamotable.write(df, file_path)



import re
from typing import Iterable
from pathlib import Path

from ..data import Data


class ParseError(RuntimeError):
    pass


# a list of commonly used base names for paths in regex form
common_name_regexes = (
    r'TS_\d+',
    r'series_\d+',
    r'_\d+',
    r'\d+',
)


def guess_name(string, name_regex=None):
    """
    guess an appropriate name based on the input string
    and a list of regexes in order of priority
    """
    if isinstance(string, Path):
        string = string.stem
    regexes = list(common_name_regexes)
    if name_regex is not None:
        regexes.insert(0, name_regex)
    for regex in regexes:
        if match := re.search(regex, str(string)):
            return match.group(0)
    else:
        return None


def listify(obj):
    """
    transform input into an appropriate list, unless already list-like
    """
    if isinstance(obj, Iterable):
        if isinstance(obj, (str, Path, Data)):
            return [obj]
        else:
            return list(obj)
    if obj is None:
        return []
    return [obj]



import dask.array as da
import mrcfile
import numpy as np
from numpy.lib.recfunctions import structured_to_unstructured

from ..data import Image
from ..utils.generic import guess_name


def read_mrc(image_path, name_regex=None, lazy=True, **kwargs):
    """
    read an mrc file
    """
    name = guess_name(image_path, name_regex)

    with mrcfile.mmap(image_path) as mrc:
        if lazy:
            data = da.from_array(mrc.data)
        else:
            data = np.asarray(mrc.data)

        # TODO: support anisotropic pixel sizes
        pixel_size = structured_to_unstructured(mrc.voxel_size)[0] or None

        stack = mrc.is_image_stack() or mrc.is_volume_stack()

    return Image(
        data=data,
        name=name,
        pixel_size=pixel_size,
        stack=stack,
    )



import pandas as pd
from scipy.spatial.transform import Rotation
import starfile

from ..utils.constants import Naaf, Relion


def write_star(particles, file_path, features=None, overwrite=False):
    """
    write particle data to disk as a .star file
    """
    df = pd.DataFrame()
    df[Relion.COORD_HEADERS] = particles.data[Naaf.COORD_HEADERS]
    rot = Rotation.concatenate(particles.data[Naaf.ROT_HEADER])
    # we use the inverse rotation in naaf
    df[Relion.EULER_HEADERS[3]] = rot.inv().as_euler(Relion.EULER, degrees=True)

    features = particles.data.drop(columns=Relion.ALL_HEADERS, errors='ignore')
    df = pd.concat([df, features], axis=1)

    if particles.pixel_size is not None:
        df[Relion.PIXEL_SIZE_HEADERS['3.1']] = particles.pixel_size

    if not str(file_path).endswith('.star'):
        file_path = str(file_path) + '.star'
    starfile.write(df, file_path, overwrite=overwrite)



import numpy as np
import pandas as pd
import starfile
from scipy.spatial.transform import Rotation

from ..data import Particles
from ..utils.constants import Naaf, Relion
from ..utils.generic import ParseError, guess_name


def extract_data(
    df,
    mode="3.1",
    name_regex=None,
    star_path="",
    **kwargs,
):
    """
    extract particle data from a starfile dataframe
    """
    if Relion.COORD_HEADERS[-1] in df.columns:
        dim = 3
    else:
        dim = 2

    if Relion.MICROGRAPH_NAME_HEADER in df.columns:
        groups = df.groupby(Relion.MICROGRAPH_NAME_HEADER)
    else:
        groups = [(star_path, df)]

    particles = []
    for micrograph_name, df_volume in groups:
        # drop global index to prevent issues with concatenation and similar
        df_volume = df_volume.reset_index(drop=True)

        name = guess_name(micrograph_name, name_regex)

        coords = np.asarray(df_volume[Relion.COORD_HEADERS[:dim]], dtype=float)
        shifts = np.asarray(
            df_volume.get(Relion.SHIFT_HEADERS[mode][:dim], 0), dtype=float
        )

        pixel_size = df_volume.get(Relion.PIXEL_SIZE_HEADERS[mode])
        if pixel_size is not None:
            pixel_size = np.asarray(pixel_size, dtype=float)
            # XXX TODO: remove the following and support variable pixel sizes?
            pixel_size = pixel_size.ravel()[0]

        # only relion 3.1 has shifts in angstroms
        if mode == "3.1":
            if pixel_size is None:
                raise ParseError("Detected Relion 3.1 format, but no pixel size data!")
            shifts = shifts / pixel_size

        coords -= shifts

        eulers = np.asarray(df_volume.get(Relion.EULER_HEADERS[dim], 0), dtype=float)
        if dim == 3:
            rot = Rotation.from_euler(Relion.EULER, eulers, degrees=True)
        else:
            rot = Rotation.from_euler(Relion.INPLANE, eulers, degrees=True)

        # we want the inverse, which when applied to basis vectors it gives us the particle orientation
        rot = rot.inv()

        features = df_volume.drop(columns=Relion.ALL_HEADERS, errors="ignore")

        data = pd.DataFrame()
        data[Naaf.COORD_HEADERS] = coords
        data[Naaf.ROT_HEADER] = np.asarray(rot)
        data = pd.concat([data, features], axis=1)

        particles.append(
            Particles(
                data=data,
                pixel_size=pixel_size,
                name=name,
            )
        )

    return particles


def parse_relion30(raw_data, **kwargs):
    """
    Attempt to parse raw data dict from starfile.read as a RELION 3.0 style star file
    """
    if len(raw_data.values()) > 1:
        raise ParseError("Cannot parse as RELION 3.0 format STAR file")

    df = list(raw_data.values())[0]
    return extract_data(df, mode="3.0", **kwargs)


def parse_relion31(raw_data, **kwargs):
    """
    Attempt to parse raw data from starfile.read as a RELION 3.1 style star file
    """
    if list(raw_data.keys()) != ["optics", "particles"]:
        raise ParseError("Cannot parse as RELION 3.1 format STAR file")

    df = raw_data["particles"].merge(raw_data["optics"])
    return extract_data(df, mode="3.1", **kwargs)


reader_functions = {
    "relion_3.0": parse_relion30,
    "relion_3.1": parse_relion31,
}


def read_star(star_path, **kwargs):
    """
    Dispatch function for reading a starfile into one or multiple ParticleBlocks
    """
    try:
        raw_data = starfile.read(star_path, always_dict=True)
    except pd.errors.EmptyDataError:  # raised sometimes by .star files with completely different data
        raise ParseError(f"the contents of {star_path} have the wrong format")

    failed_reader_functions = []
    for style, reader_function in reader_functions.items():
        try:
            particle_blocks = reader_function(raw_data, star_path=star_path, **kwargs)
            return particle_blocks
        except ParseError:
            failed_reader_functions.append((style, reader_function))
    raise ParseError(f"Failed to parse {star_path} using {failed_reader_functions}")



import emfile


def write_em(image, file_path, overwrite=False):
    """
    write an image to disk as an .em file
    """
    if not file_path.endswith('.em'):
        file_path = file_path + '.em'
    # TODO: pass pixel size and other stuff to header
    emfile.write(file_path, image.data, header_params={}, overwrite=overwrite)



from scipy.spatial.transform import Rotation
import starfile

from ..utils.generic import guess_name
from ..utils.constants import Relion


def read_cbox(
    cbox_path,
    name_regex=None,
    **kwargs,
):
    data = starfile.read(cbox_path)['cryolo']
    coords = data[[f'Coordinate{axis}' for axis in 'XYZ']].to_numpy()
    rot = Rotation.identity(len(coords)).as_euler(Relion.EULER)
    name = guess_name(cbox_path, name_regex)
    return coords, rot, {'name': name}



import numpy as np
import pandas as pd
import mrcfile
import starfile

from naaf.reading.main import read


def test_read_path(tmp_path):
    # mrc file
    mrc_path1 = tmp_path / 'test1.mrc'
    mrc_path2 = tmp_path / 'test2.mrc'
    mrcfile.new(str(mrc_path1), np.ones((3, 3), dtype=np.float32))
    mrcfile.new(str(mrc_path2), np.ones((3, 3), dtype=np.float32))

    # star file
    df = pd.DataFrame({
        'rlnCoordinateX': [1, 2],
        'rlnCoordinateY': [1, 2],
        'rlnCoordinateZ': [1, 2],
        'rlnOriginX': [1, 2],
        'rlnOriginY': [1, 2],
        'rlnOriginZ': [1, 2],
        'rlnAngleRot': [1, 2],
        'rlnAngleTilt': [1, 2],
        'rlnAnglePsi': [1, 2],
        'rlnMicrographName': ['test1', 'test2'],
    })
    star_path = tmp_path / 'test.star'
    starfile.write(df, star_path)

    data = read(tmp_path, name_regex=r'test\d')

    assert len(data) == 4

    sources = [d.source for d in data]
    for s in (mrc_path1, mrc_path2, star_path):
        assert s in sources



import numpy as np
from scipy.spatial.transform import Rotation

from ..utils.generic import guess_name
from ..utils.constants import Relion
from ..data import Particles


def read_box(
    path,
    name_regex=None,
    **kwargs,
):
    coords = np.loadtxt(path)
    rot = Rotation.identity(len(coords)).as_euler(Relion.EULER)
    name = guess_name(path, name_regex)

    return Particles(
        coords=coords,
        rot=rot,
        name=name,
    )



import numpy as np
import pandas as pd
import dask.array as da
from scipy.spatial.transform import Rotation


def assert_data_equal(data1, data2):
    d1 = data1.dict()
    d2 = data2.dict()
    assert d1.keys() == d2.keys(), 'models do not have the same set of fields'
    for field in d1.keys():
        if isinstance(d1[field], (np.ndarray, da.Array)):
            try:
                if not np.allclose(d1[field], d2[field]):
                    raise AssertionError(f'numeric array fields "{field}" are not within error')
            except TypeError:
                if not np.all(d1[field] == d2[field]):
                    raise AssertionError(f'array fields "{field}" are not equal')
        elif isinstance(d1[field], pd.DataFrame):
            assert set(d1[field].columns) == set(d2[field].columns), f'dataframe fields "{field}" have different columns'
            for col in d1[field]:
                if isinstance(d1[field][col][0], Rotation):
                    s_rot = Rotation.concatenate(d1[field][col])
                    o_rot = Rotation.concatenate(d2[field][col])
                    # invert and multiply should give zero
                    back_forth = (s_rot * o_rot.inv()).reduce().as_rotvec()
                    if not np.allclose(back_forth, 0):
                        raise AssertionError(f'Rotation columns "{col}" of field "{field}" are not equivalent')
                else:
                    try:
                        if not np.allclose(d1[field][col], d2[field][col]):
                            raise AssertionError(f'numeric array columns "{col}" of field "{field}" are not within error')
                    except TypeError:
                        if not np.all(d1[field][col] == d2[field][col]):
                            raise AssertionError(f'array columns "{col}" of field "{field}" are not equal')
        else:
            if not d1[field] == d2[field]:
                return False
    return True



import numpy as np

from naaf.writing.em import write_em
from naaf.data import Image


def test_write_em(tmp_path):
    file_path = tmp_path / 'test.em'
    imageblock = Image(data=np.ones((3, 3, 3)))
    write_em(imageblock, str(file_path))



import numpy as np
import dask.array as da
import emfile

from ..utils.generic import guess_name
from ..data import Image


def read_em(
    image_path,
    name_regex=None,
    lazy=True,
    **kwargs
):
    """
    read an em image file
    """
    name = guess_name(image_path, name_regex)

    header, data = emfile.read(image_path, mmap=True)

    if lazy:
        data = da.from_array(data)
    else:
        data = np.asarray(data)

    pixel_size = header['OBJ'] or None

    return Image(
        data=data,
        name=name,
        pixel_size=pixel_size,
    )



import numpy as np

from naaf.writing.mrc import write_mrc
from naaf.data import Image


def test_write_mrc(tmp_path):
    file_path = tmp_path / 'test.mrc'
    imageblock = Image(data=np.ones((3, 3, 3), dtype=np.float32))
    write_mrc(imageblock, str(file_path))
